/* sane - Scanner Access Now Easy.
   Copyright (C) 2002-2003 Martin Palm
   most based on code Copyright (C) 1999 Juergen G. Schimmer
   and on code from xawtv etc. Thanks for this to Gerd Knorr
   
   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   This file implements a SANE backend for v4l-Devices.
*/

#define BUILD 3

#include "../include/sane/config.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/mman.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "../include/sane/sane.h"
#include "../include/sane/sanei.h"
#include "../include/sane/saneopts.h"

#include <sys/ioctl.h>
#include <asm/types.h>		/* XXX glibc */
#include <linux/videodev.h>

#include "v4l.h"

#define BACKEND_NAME v4l
#define CONFIG_FILE  "/etc/v4lchannel.conf"
#include "../include/sane/sanei_backend.h"

#ifndef PATH_MAX
#define PATH_MAX    1024
#endif

#include "../include/sane/sanei_config.h"
#define V4L_CONFIG_FILE "v4l.conf"
#define UP    1
#define DOWN -1
#define MAX(a,b) a>b?a:b
#define MIN(a,b) a>b?b:a
#define NO_CHANNELLIST "No Channellist available"

/*----------------------------------------------------------------------------*/
/* TODO:                                                                      */
/* devnumber - compare in attach                                              */
/* fixed-point arithmetric                                                    */
/*----------------------------------------------------------------------------*/
static const SANE_Device **devlist = NULL;
static SANE_Int num_devices;
static V4L_Device *first_dev;
static V4L_Scanner *first_handle;

static const SANE_Int color_depth_list[] = {
  3,				/* # of elements */
  8, 16, 24
};

static SANE_String Normlist[] = {
  "PAL", "NTSC", "SECAM", 0
};

static const SANE_Range percent_range = {
  0, SANE_FIX (100.0), SANE_FIX (0.0625)
};
static const SANE_Range ft_range = { -64, 64, 0 };	/* +/- 4 MHz */
static const SANE_Range x_range[3] = {
  {0, 767, 0}, {0, 639, 0}, {0, 767, 0}
};
static const SANE_Range y_range[3] = {
  {0, 575, 0}, {0, 479, 0}, {0, 575, 0}
};

/*----------------------------------------------------------------------------*/
#if 0
static SANE_Bool
v4l_is_little_endian (void)
{
  SANE_Int testval = 255;
  unsigned char *firstbyte = (unsigned char *) &testval;

  if (255 == *firstbyte)
    return SANE_TRUE;
  return SANE_FALSE;
}
#endif
/*----------------------------------------------------------------------------*/
static SANE_Status
v4l_read_channelfile (V4L_Scanner * s)
{
  char string[PATH_MAX];
  SANE_String cpa;
  FILE *fp;
  struct channellist_t *newentry;
  struct channellist_t *listbase = NULL;
  double freq;
  char chname[40];
  SANE_Int count;

  /* don't initialize more than once */
  if (s->NumChannels || s->Chnamearray || s->Chfreqarray)
    {
      DBG (2, "v4l_read channelfile: reinit not allowed\n");
      return SANE_STATUS_INVAL;
    }
  if (!s->Scalefakt)
    return SANE_STATUS_INVAL;
  fp = fopen (CONFIG_FILE, "r");
  if (fp)
    {
      while (fgets (string, PATH_MAX, fp))
	{
	  if (*string == '#')
	    continue;		/* ignore line comments */
	  /* Remove leading space and trailing comments */
	  for (cpa = string; *cpa && isspace (*cpa); ++cpa);
	  if (!strlen (cpa))
	    continue;
	  /* Remove trailing space and trailing comments (TODO) */
	  /* read values from string */
	  if (sscanf (cpa, "%s %lf", &chname[0], &freq) != 2)
	    continue;
	  if (!strlen (chname))
	    continue;
	  if (freq < (double) s->MinFrequency / (double) s->Scalefakt)
	    continue;
	  if (freq > (double) s->MaxFrequency / (double) s->Scalefakt)
	    continue;
	  /* put values in a list */
	  newentry =
	    (struct channellist_t *) malloc (sizeof (struct channellist_t));
	  if (!newentry)
	    return SANE_STATUS_NO_MEM;
	  newentry->next = listbase;
	  newentry->freq = (unsigned long) (freq * 100.);
	  newentry->chname = strdup (chname);
	  if (!strlen (newentry->chname))
	    {
	      free (newentry);
	    }
	  else
	    {
	      listbase = newentry;
	      s->NumChannels = s->NumChannels + 1;
	    }
	}
      fclose (fp);
    }

  /* now convert the list to an array */
  if (!s->NumChannels)
    return SANE_STATUS_GOOD;
  s->Chnamearray =
    (SANE_String *) malloc ((s->NumChannels + 1) * sizeof (SANE_String));
  if (!s->Chnamearray)
    {
      s->NumChannels = 0;
      return SANE_STATUS_NO_MEM;
    }
  s->Chfreqarray = (unsigned long *) malloc (s->NumChannels * sizeof (int));
  if (!s->Chfreqarray)
    {
      s->NumChannels = 0;
      free (s->Chnamearray);
      return SANE_STATUS_NO_MEM;
    }
  for (count = s->NumChannels - 1; listbase; count--)
    {
      s->Chfreqarray[count] = listbase->freq;
      s->Chnamearray[count] = strdup (listbase->chname);
      if (!s->Chnamearray[count])
	return SANE_STATUS_NO_MEM;
      free (listbase->chname);
      newentry = listbase->next;
      free (listbase);
      listbase = newentry;
    }
  s->Chnamearray[s->NumChannels] = (SANE_String) NULL;
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
/* Get OptionIndex from Optionstring                                          */
/*----------------------------------------------------------------------------*/
static SANE_Int
v4l_get_option_int (const void *val, const SANE_String * list,
		    const SANE_Int def)
{
  SANE_Int i;

  if (!val || !list)
    return def;
  for (i = 0; list[i]; i++)
    {
      if (!strcmp (val, list[i]))
	return i;
    }
  return def;			/* string not found */
}

/*----------------------------------------------------------------------------*/
/* Check if current channel has a tuner                                       */
/*----------------------------------------------------------------------------*/
static SANE_Bool
v4l_has_tuner (const V4L_Scanner * s)
{
  struct video_channel input;

  if (!s)
    return SANE_FALSE;
  memset (&input, 0, sizeof (struct video_channel));
  input.channel =
    v4l_get_option_int (s->val[OPT_INPUT].s, s->InputNameArray, 1);
  if (ioctl (s->fd, VIDIOCGCHAN, &input) < 0)
    {
      DBG (2, "v4l_has_tuner: ioctl VIDICGCHAN failed\n");
      return SANE_FALSE;
    }
  if (!(input.flags & VIDEO_VC_TUNER))
    return SANE_FALSE;
  return SANE_TRUE;
}

/*----------------------------------------------------------------------------*/
/* Tune to channel                                                            */
/*----------------------------------------------------------------------------*/
static SANE_Status
v4l_tune_channel (V4L_Scanner * s)
{
  unsigned long rfreq;

  if (!v4l_has_tuner (s) || !s->NumChannels)
    return SANE_STATUS_GOOD;
  if (s->val[OPT_PRESET].w < 0)
    s->val[OPT_PRESET].w = 0;
  if (s->val[OPT_PRESET].w >= s->NumChannels)
    s->val[OPT_PRESET].w = s->NumChannels - 1;
  rfreq = s->Chfreqarray[s->val[OPT_PRESET].w] * s->Scalefakt / 100;
  rfreq = rfreq + (SANE_Int) s->val[OPT_FINETUNE].w * s->Scalefakt / 16;
  if (ioctl (s->fd, VIDIOCSFREQ, &rfreq) < 0)
    {
      DBG (2, "v4l_tune_channel: ioctl VIDICSFREQ failed\n");
      return SANE_STATUS_IO_ERROR;
    }
  /* Show channel to user */
  if (s->val[OPT_CHANNEL].s)
    free (s->val[OPT_CHANNEL].s);
  s->val[OPT_CHANNEL].s = strdup (s->Chnamearray[s->val[OPT_PRESET].w]);
  if (!s->val[OPT_CHANNEL].s)
    return SANE_STATUS_NO_MEM;
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
/* Step one channel up or down                                                */
/*----------------------------------------------------------------------------*/
static SANE_Status
v4l_step_channel (V4L_Scanner * s, const SANE_Int UpDown)
{
  if (!v4l_has_tuner (s) || !s->NumChannels)
    return SANE_STATUS_GOOD;
  if (UP == UpDown)
    {
      s->val[OPT_PRESET].w = s->val[OPT_PRESET].w + 1;
      if (s->val[OPT_PRESET].w >= s->NumChannels)
	s->val[OPT_PRESET].w = 0;
    }
  else
    {
      s->val[OPT_PRESET].w = s->val[OPT_PRESET].w - 1;
      if (s->val[OPT_PRESET].w < 0)
	s->val[OPT_PRESET].w = s->val[OPT_PRESET].w - 1;
    }
  return v4l_tune_channel (s);
}

/*----------------------------------------------------------------------------*/
/* check if current input is Mutable                                          */
/*----------------------------------------------------------------------------*/
static SANE_Bool
v4l_is_mutable (const V4L_Scanner * s, const SANE_Int channel)
{
  struct video_channel input;
  struct video_audio audio;

  memset (&input, 0, sizeof (struct video_channel));
  input.channel = channel;
  if (ioctl (s->fd, VIDIOCGCHAN, &input) < 0)
    {
      DBG (2, "v4l_is_mutable: ioctl VIDICGCHAN failed\n");
      return SANE_FALSE;
    }
  memset (&audio, 0, sizeof (struct video_audio));
  audio.audio = channel;
  if (ioctl (s->fd, VIDIOCGAUDIO, &audio) < 0)
    {
      DBG (2, "v4l_is_mutable: ioctl VIDICGAUDIO failed\n");
      return SANE_FALSE;
    }
  if ((input.flags & VIDEO_VC_AUDIO) && (audio.flags & VIDEO_AUDIO_MUTABLE))
    return SANE_TRUE;
  return SANE_FALSE;
}

/*----------------------------------------------------------------------------*/
/* Mute current Input                                                         */
/*----------------------------------------------------------------------------*/
static SANE_Status
v4l_mute (const V4L_Scanner * s, const SANE_Bool On)
{
  struct video_audio audio;
  SANE_Int channel;

  channel = v4l_get_option_int (s->val[OPT_INPUT].s, s->InputNameArray, 1);
  if (!v4l_is_mutable (s, channel))
    return SANE_STATUS_GOOD;
  memset (&audio, 0, sizeof (struct video_audio));
  audio.audio = channel;
  if (ioctl (s->fd, VIDIOCGAUDIO, &audio) < 0)
    {
      DBG (2, "v4l_mute: ioctl VIDICGAUDIO failed\n");
      return SANE_STATUS_IO_ERROR;
    }
  if (On)
    {
      audio.flags |= VIDEO_AUDIO_MUTE;
    }
  else
    {
      audio.flags &= ~VIDEO_AUDIO_MUTE;
    }
  if (ioctl (s->fd, VIDIOCSAUDIO, &audio) < 0)
    {
      DBG (2, "v4l_is_mutable: ioctl VIDICGAUDIO failed\n");
      return SANE_STATUS_IO_ERROR;
    }
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
/* selects video input                                                        */
/*----------------------------------------------------------------------------*/
static SANE_Status
v4l_set_input (V4L_Scanner * s)
{
  struct video_channel input;
  struct video_tuner tuner;
  SANE_Int count;
  SANE_Int i;

  memset (&input, 0x00, sizeof (input));
  input.channel =
    v4l_get_option_int (s->val[OPT_INPUT].s, s->InputNameArray, 1);
  input.norm = v4l_get_option_int (s->val[OPT_NORM].s, Normlist, 0);
  s->opt[OPT_TL_X].constraint.range = &x_range[input.norm];
  s->opt[OPT_TL_Y].constraint.range = &y_range[input.norm];
  s->opt[OPT_BR_X].constraint.range = &x_range[input.norm];
  s->opt[OPT_BR_Y].constraint.range = &y_range[input.norm];
  s->val[OPT_TL_X].w = MAX (s->val[OPT_TL_X].w, x_range[input.norm].min);
  s->val[OPT_TL_Y].w = MAX (s->val[OPT_TL_Y].w, y_range[input.norm].min);
  s->val[OPT_BR_X].w = MAX (s->val[OPT_BR_X].w, x_range[input.norm].max);
  s->val[OPT_BR_Y].w = MAX (s->val[OPT_BR_Y].w, y_range[input.norm].max);
  if (ioctl (s->fd, VIDIOCSCHAN, &input) < 0)
    {
      DBG (2, "v4l_set_input: ioctl VIDICSCHAN(1) failed\n");
      return SANE_STATUS_IO_ERROR;
    }
  /* If we have selected the input connected to the tuner */
  /* it is certainly a good idea to switch the Videonorm  */
  /* to a value supported by the tuner */
  if (ioctl (s->fd, VIDIOCGTUNER, &tuner) >= 0 &&
      ioctl (s->fd, VIDIOCGCHAN, &input) >= 0)
    {
      if ((input.flags & VIDEO_VC_TUNER) &&
	  (!(tuner.flags & VIDEO_TUNER_NORM)))
	{
	  /* check for multinorm tuner */
	  count = 0;
	  for (i = 0; i < 3; i++)
	    {
	      if (tuner.flags & (2 ^ i))
		count++;
	    }
	  /* do not change norm for multinorm tuners */
	  if (count < 2)
	    {
	      input.norm = VIDEO_MODE_PAL;
	      if (tuner.flags & VIDEO_TUNER_NTSC)
		input.norm = VIDEO_MODE_NTSC;
	      else if (tuner.flags & VIDEO_TUNER_SECAM)
		input.norm = VIDEO_MODE_SECAM;
	      if (s->val[OPT_NORM].s)
		free (s->val[OPT_NORM].s);
	      s->val[OPT_NORM].s = strdup (Normlist[input.norm]);
	      if (!s->val[OPT_NORM].s)
		return SANE_STATUS_NO_MEM;
	      if (ioctl (s->fd, VIDIOCSCHAN, &input) < 0)
		{
		  DBG (2, "v4l_set_input: ioctl VIDICSCHAN(2) failed\n");
		  return SANE_STATUS_IO_ERROR;
		}
	    }
	  /* Tune the current channel */
	  if (s->NumChannels)
	    {
	      v4l_tune_channel (s);
	      /* enable tuning controls */
	      s->opt[OPT_FINETUNE].cap &= ~SANE_CAP_INACTIVE;
	      s->opt[OPT_CHANNEL].cap &= ~SANE_CAP_INACTIVE;
	      s->opt[OPT_CHANNEL_UP].cap &= ~SANE_CAP_INACTIVE;
	      s->opt[OPT_CHANNEL_DOWN].cap &= ~SANE_CAP_INACTIVE;
	    }
	  else
	    {
	      if (s->val[OPT_CHANNEL].s)
		free (s->val[OPT_CHANNEL].s);
	      s->val[OPT_CHANNEL].s = strdup (NO_CHANNELLIST);
	      if (!s->val[OPT_CHANNEL].s)
		return SANE_STATUS_NO_MEM;
	    }
	}
      else
	{
	  /* disable tuning controls */
	  s->opt[OPT_FINETUNE].cap |= SANE_CAP_INACTIVE;
	  s->opt[OPT_CHANNEL].cap |= SANE_CAP_INACTIVE;
	  s->opt[OPT_CHANNEL_UP].cap |= SANE_CAP_INACTIVE;
	  s->opt[OPT_CHANNEL_DOWN].cap |= SANE_CAP_INACTIVE;
	}
    }
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
/* get input names for option struct                                          */
/*----------------------------------------------------------------------------*/
static SANE_Status
v4l_get_inputnames (V4L_Scanner * s, const SANE_Int inputcount)
{
  struct video_channel input;
  struct video_tuner tuner;
  SANE_Int i;

  /* Read all available inputnames, build struct for options */
  s->InputNameArray =
    (SANE_String *) malloc ((inputcount + 1) * sizeof (SANE_String));
  for (i = 0; i < inputcount; i++)
    {
      memset (&input, 0, sizeof (input));
      input.channel = i;
      if (ioctl (s->fd, VIDIOCGCHAN, &input) < 0)
	{
	  DBG (2, "v4l_get_inputnames: ioctl VIDICGCHAN failed\n");
	  return SANE_STATUS_IO_ERROR;
	}
      s->InputNameArray[i] = strdup (input.name);
      if (!s->InputNameArray[i])
	return SANE_STATUS_NO_MEM;
    }
  /* last one for end */
  s->InputNameArray[i] = NULL;
  /* get tuner info */
  s->Scalefakt = 0;
  s->MinFrequency = 0;
  s->MaxFrequency = 0;
  if (ioctl (s->fd, VIDIOCGTUNER, &tuner) >= 0)
    {
      if (tuner.flags & VIDEO_TUNER_LOW)
	{
	  s->Scalefakt = 16000;
	}
      else
	{
	  s->Scalefakt = 16;
	}
      s->MinFrequency = tuner.rangelow;
      s->MaxFrequency = tuner.rangehigh;
    }
  else
    {
      DBG (2, "v4l_get_inputnames: ioctl VIDICGTUNER failed\n");
    }
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
static SANE_Fixed
v4l_scale_int2fix_percent (SANE_Word val)
{
  return SANE_FIX ((double) val / 65535. * SANE_UNFIX (percent_range.max));
}

/*----------------------------------------------------------------------------*/
static SANE_Int
v4l_scale_fix_percent2int (SANE_Fixed val)
{
  return (SANE_Int) (SANE_UNFIX (val) * 65535. /
		     SANE_UNFIX (percent_range.max));
}

/*----------------------------------------------------------------------------*/
static SANE_Int
v4l_init_depth (SANE_Int depth)
{
  if (depth >= 24)
    return 24;
  if (depth > 8)
    return 16;
  return 8;
}

/*----------------------------------------------------------------------------*/
static SANE_Status
v4l_init_options (V4L_Scanner * s, SANE_Int numchannels)
{
  SANE_Int i;
  struct video_picture pict;
  SANE_Status status;

  if (ioctl (s->fd, VIDIOCGPICT, &pict) < 0)
    {
      DBG (1, "v4l_init_options: Can not get Picture Parameters\n");
      return SANE_STATUS_IO_ERROR;
    }

  status = v4l_get_inputnames (s, numchannels);
  if (SANE_STATUS_GOOD != status)
    return status;

  status = v4l_read_channelfile (s);
  if (SANE_STATUS_GOOD != status)
    return status;

  for (i = 0; i < NUM_OPTIONS; ++i)
    {
      s->opt[i].size = sizeof (SANE_Word);
      s->opt[i].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT |
	SANE_CAP_ALWAYS_SETTABLE;
    }

  s->opt[OPT_NUM_OPTS].title = SANE_TITLE_NUM_OPTIONS;
  s->opt[OPT_NUM_OPTS].desc = SANE_DESC_NUM_OPTIONS;
  s->opt[OPT_NUM_OPTS].type = SANE_TYPE_INT;
  s->opt[OPT_NUM_OPTS].cap = SANE_CAP_SOFT_DETECT;
  s->val[OPT_NUM_OPTS].w = NUM_OPTIONS;

  /* "Mode" group: */
  s->opt[OPT_MODE_GROUP].title = "Scan Mode";
  s->opt[OPT_MODE_GROUP].desc = "";
  s->opt[OPT_MODE_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_MODE_GROUP].cap = 0;
  s->opt[OPT_MODE_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

  /* input source */
  s->opt[OPT_INPUT].name = SANE_NAME_SCAN_SOURCE;
  s->opt[OPT_INPUT].title = "Input Source";
  s->opt[OPT_INPUT].desc =
    "Selects the input of the v4l device (e.g. Television" " or Composite)";
  s->opt[OPT_INPUT].type = SANE_TYPE_STRING;
  s->opt[OPT_INPUT].unit = SANE_UNIT_NONE;
  if (s->InputNameArray)
    {
      s->opt[OPT_INPUT].constraint_type = SANE_CONSTRAINT_STRING_LIST;
      s->opt[OPT_INPUT].constraint.string_list =
	(SANE_String_Const *) s->InputNameArray;
      s->val[OPT_INPUT].s = strdup (s->InputNameArray[1]);
      if (!s->val[OPT_INPUT].s)
	return SANE_STATUS_NO_MEM;
      status = v4l_set_input (s);
      if (SANE_STATUS_GOOD != status)
	return status;
    }
  else
    {
      s->opt[OPT_INPUT].constraint_type = SANE_CONSTRAINT_NONE;
      s->val[OPT_INPUT].s = strdup ("No Inputs found");
      s->opt[OPT_INPUT].cap |= SANE_CAP_INACTIVE;
    }

  /* video_norm */
  s->opt[OPT_NORM].name = "videonorm";
  s->opt[OPT_NORM].title = "Video Norm";
  s->opt[OPT_NORM].desc = "Selects the norm used for capturing the Image";
  s->opt[OPT_NORM].type = SANE_TYPE_STRING;
  s->opt[OPT_NORM].unit = SANE_UNIT_NONE;
  s->opt[OPT_NORM].constraint_type = SANE_CONSTRAINT_STRING_LIST;
  s->opt[OPT_NORM].constraint.string_list = (SANE_String_Const *) Normlist;
  s->val[OPT_NORM].s = strdup (Normlist[0]);
  if (!s->val[OPT_NORM].s)
    return SANE_STATUS_NO_MEM;

  /* bit-depth */
  s->opt[OPT_DEPTH].name = SANE_NAME_BIT_DEPTH;
  s->opt[OPT_DEPTH].title = "Pixel depth";
  s->opt[OPT_DEPTH].desc = "Number of bits per pixel.";
  s->opt[OPT_DEPTH].type = SANE_TYPE_INT;
  s->opt[OPT_DEPTH].unit = SANE_UNIT_BIT;
  s->opt[OPT_DEPTH].constraint_type = SANE_CONSTRAINT_WORD_LIST;
  s->opt[OPT_DEPTH].constraint.word_list = color_depth_list;
  s->val[OPT_DEPTH].w = v4l_init_depth (pict.depth);

  /* "Geometry" group: */
  s->opt[OPT_GEOMETRY_GROUP].title = "Geometry";
  s->opt[OPT_GEOMETRY_GROUP].desc = "";
  s->opt[OPT_GEOMETRY_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_GEOMETRY_GROUP].cap = SANE_CAP_ADVANCED;
  s->opt[OPT_GEOMETRY_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

  /* top-left x */
  s->opt[OPT_TL_X].name = SANE_NAME_SCAN_TL_X;
  s->opt[OPT_TL_X].title = SANE_TITLE_SCAN_TL_X;
  s->opt[OPT_TL_X].desc = SANE_DESC_SCAN_TL_X;
  s->opt[OPT_TL_X].type = SANE_TYPE_INT;
  s->opt[OPT_TL_X].unit = SANE_UNIT_PIXEL;
  s->opt[OPT_TL_X].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_TL_X].constraint.range = &x_range[0];
  s->val[OPT_TL_X].w = x_range[0].min;

  /* top-left y */
  s->opt[OPT_TL_Y].name = SANE_NAME_SCAN_TL_Y;
  s->opt[OPT_TL_Y].title = SANE_TITLE_SCAN_TL_Y;
  s->opt[OPT_TL_Y].desc = SANE_DESC_SCAN_TL_Y;
  s->opt[OPT_TL_Y].type = SANE_TYPE_INT;
  s->opt[OPT_TL_Y].unit = SANE_UNIT_PIXEL;
  s->opt[OPT_TL_Y].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_TL_Y].constraint.range = &y_range[0];
  s->val[OPT_TL_Y].w = y_range[0].min;

  /* bottom-right x */
  s->opt[OPT_BR_X].name = SANE_NAME_SCAN_BR_X;
  s->opt[OPT_BR_X].title = SANE_TITLE_SCAN_BR_X;
  s->opt[OPT_BR_X].desc = SANE_DESC_SCAN_BR_X;
  s->opt[OPT_BR_X].type = SANE_TYPE_INT;
  s->opt[OPT_BR_X].unit = SANE_UNIT_PIXEL;
  s->opt[OPT_BR_X].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BR_X].constraint.range = &x_range[0];
  s->val[OPT_BR_X].w = x_range[0].max;

  /* bottom-right y */
  s->opt[OPT_BR_Y].name = SANE_NAME_SCAN_BR_Y;
  s->opt[OPT_BR_Y].title = SANE_TITLE_SCAN_BR_Y;
  s->opt[OPT_BR_Y].desc = SANE_DESC_SCAN_BR_Y;
  s->opt[OPT_BR_Y].type = SANE_TYPE_INT;
  s->opt[OPT_BR_Y].unit = SANE_UNIT_PIXEL;
  s->opt[OPT_BR_Y].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BR_Y].constraint.range = &y_range[0];
  s->val[OPT_BR_Y].w = y_range[0].max;

  /* "Special" group: */
  s->opt[OPT_SPECIAL_GROUP].title = "Special";
  s->opt[OPT_SPECIAL_GROUP].desc = "";
  s->opt[OPT_SPECIAL_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_SPECIAL_GROUP].cap = SANE_CAP_ADVANCED;
  s->opt[OPT_SPECIAL_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

  /* audio */
  s->opt[OPT_AUDIO].name = "audio";
  s->opt[OPT_AUDIO].title = "Audio On/Off";
  s->opt[OPT_AUDIO].desc = "Switches Audio On/Off (not captured)";
  s->opt[OPT_AUDIO].type = SANE_TYPE_BOOL;
  s->opt[OPT_AUDIO].size = sizeof (SANE_Word);
  s->opt[OPT_AUDIO].constraint_type = SANE_CONSTRAINT_NONE;
  s->val[OPT_AUDIO].w = 0;
  if (!v4l_is_mutable
      (s, v4l_get_option_int (s->val[OPT_INPUT].s, s->InputNameArray, 1)))
    {
      s->opt[OPT_AUDIO].cap |= SANE_CAP_INACTIVE;
    }

  /* preset station number */
  s->opt[OPT_PRESET].name = "chpreset";
  s->opt[OPT_PRESET].title = "Channel Preset";
  s->opt[OPT_PRESET].desc = "Preset Number";
  s->opt[OPT_PRESET].type = SANE_TYPE_INT;
  s->opt[OPT_PRESET].size = sizeof (SANE_Word);
  s->opt[OPT_PRESET].unit = SANE_UNIT_NONE;
  s->opt[OPT_PRESET].constraint_type = SANE_CONSTRAINT_NONE;
  s->val[OPT_PRESET].w = 0;

  /* channel */
  s->opt[OPT_CHANNEL].name = "station";
  s->opt[OPT_CHANNEL].title = "Channel/Station";
  s->opt[OPT_CHANNEL].desc = "Selects channel/station to capture";
  s->opt[OPT_CHANNEL].type = SANE_TYPE_STRING;
  s->opt[OPT_CHANNEL].unit = SANE_UNIT_NONE;
  s->opt[OPT_CHANNEL].constraint_type = SANE_CONSTRAINT_NONE;
  if (s->Chnamearray)
    {
      s->val[OPT_CHANNEL].s = strdup (s->Chnamearray[s->val[OPT_PRESET].w]);
    }
  else
    {
      s->val[OPT_CHANNEL].s = strdup (NO_CHANNELLIST);
    }
  if (!s->val[OPT_CHANNEL].s)
    return SANE_STATUS_NO_MEM;
  s->opt[OPT_CHANNEL].cap = SANE_CAP_HARD_SELECT | SANE_CAP_INACTIVE;

  /* finetune */
  s->opt[OPT_FINETUNE].name = "finetune";
  s->opt[OPT_FINETUNE].title = "Fine Tuning";
  s->opt[OPT_FINETUNE].desc = "Adjusts the selected channel";
  s->opt[OPT_FINETUNE].type = SANE_TYPE_INT;
  s->opt[OPT_FINETUNE].unit = SANE_UNIT_NONE;
  if (!s->NumChannels)
    s->opt[OPT_FINETUNE].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_FINETUNE].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_FINETUNE].constraint.range = &ft_range;
  s->val[OPT_FINETUNE].w = 0;

  /* channel up */
  s->opt[OPT_CHANNEL_UP].name = "tuneup";
  s->opt[OPT_CHANNEL_UP].title = "Channel/Station Up";
  s->opt[OPT_CHANNEL_UP].desc = "Tunes one channel/station up";
  s->opt[OPT_CHANNEL_UP].type = SANE_TYPE_BUTTON;
  s->opt[OPT_CHANNEL_UP].unit = SANE_UNIT_NONE;
  if (!s->NumChannels)
    s->opt[OPT_CHANNEL_UP].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_CHANNEL_UP].constraint_type = SANE_CONSTRAINT_NONE;
  s->val[OPT_CHANNEL_UP].w = SANE_FALSE;

  /* channel down */
  s->opt[OPT_CHANNEL_DOWN].name = "tunedown";
  s->opt[OPT_CHANNEL_DOWN].title = "Channel/Station Down";
  s->opt[OPT_CHANNEL_DOWN].desc = "Tunes one channel/station down";
  s->opt[OPT_CHANNEL_DOWN].type = SANE_TYPE_BUTTON;
  s->opt[OPT_CHANNEL_DOWN].unit = SANE_UNIT_NONE;
  if (!s->NumChannels)
    s->opt[OPT_CHANNEL_DOWN].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_CHANNEL_DOWN].constraint_type = SANE_CONSTRAINT_NONE;
  s->val[OPT_CHANNEL_DOWN].w = SANE_FALSE;

  /* "Enhancement" group: */
  s->opt[OPT_ENHANCEMENT_GROUP].title = "Enhancement";
  s->opt[OPT_ENHANCEMENT_GROUP].desc = "";
  s->opt[OPT_ENHANCEMENT_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_ENHANCEMENT_GROUP].cap = 0;
  s->opt[OPT_ENHANCEMENT_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

  /* brightness */
  s->opt[OPT_BRIGHTNESS].name = SANE_NAME_BRIGHTNESS;
  s->opt[OPT_BRIGHTNESS].title = SANE_TITLE_BRIGHTNESS;
  s->opt[OPT_BRIGHTNESS].desc = SANE_DESC_BRIGHTNESS
    "This controls the brightness of the Picture returned from"
    "the v4l device.";
  s->opt[OPT_BRIGHTNESS].type = SANE_TYPE_FIXED;
  s->opt[OPT_BRIGHTNESS].size = sizeof (SANE_Word);
  s->opt[OPT_BRIGHTNESS].unit = SANE_UNIT_PERCENT;
  s->opt[OPT_BRIGHTNESS].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BRIGHTNESS].constraint.range = &percent_range;
  s->val[OPT_BRIGHTNESS].w = v4l_scale_int2fix_percent (pict.brightness);

  /* contrast */
  s->opt[OPT_CONTRAST].name = SANE_NAME_CONTRAST;
  s->opt[OPT_CONTRAST].title = SANE_TITLE_CONTRAST;
  s->opt[OPT_CONTRAST].desc = SANE_DESC_CONTRAST
    " This controls the contrast of the Picture returned from"
    "the v4l device";
  s->opt[OPT_CONTRAST].type = SANE_TYPE_FIXED;
  s->opt[OPT_CONTRAST].size = sizeof (SANE_Word);
  s->opt[OPT_CONTRAST].unit = SANE_UNIT_PERCENT;
  s->opt[OPT_CONTRAST].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_CONTRAST].constraint.range = &percent_range;
  s->val[OPT_CONTRAST].w = v4l_scale_int2fix_percent (pict.contrast);

  /* hue */
  s->opt[OPT_HUE].name = SANE_NAME_HUE;
  s->opt[OPT_HUE].title = SANE_TITLE_HUE;
  s->opt[OPT_HUE].desc = SANE_DESC_HUE
    "This value should be selected so that pink looks realy pink";
  s->opt[OPT_HUE].type = SANE_TYPE_FIXED;
  s->opt[OPT_HUE].size = sizeof (SANE_Word);
  s->opt[OPT_HUE].unit = SANE_UNIT_PERCENT;
  s->opt[OPT_HUE].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_HUE].constraint.range = &percent_range;
  s->val[OPT_HUE].w = v4l_scale_int2fix_percent (pict.hue);

  /* saturation */
  s->opt[OPT_SATURATION].name = SANE_NAME_SATURATION;
  s->opt[OPT_SATURATION].title = SANE_TITLE_SATURATION;
  s->opt[OPT_SATURATION].desc = SANE_DESC_SATURATION
    "This value controlls wether the Picture looks Black and White"
    "or shows to much color";
  s->opt[OPT_SATURATION].type = SANE_TYPE_FIXED;
  s->opt[OPT_SATURATION].size = sizeof (SANE_Word);
  s->opt[OPT_SATURATION].unit = SANE_UNIT_PERCENT;
  s->opt[OPT_SATURATION].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_SATURATION].constraint.range = &percent_range;
  s->val[OPT_SATURATION].w = v4l_scale_int2fix_percent (pict.colour);

  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
static SANE_Status
v4l_attach (const char *devname)
{
  V4L_Device *dev;
  struct video_capability capability;
  SANE_Int v4lfd;

  /* don't attach twice */
  for (dev = first_dev; dev; dev = dev->next)
    {
      /* just comparing devnames doesn't allways do the trick */
      /* compare devnumbers will do it */
      if (!strcmp (dev->sane.name, devname))
	{
	  DBG (5, "v4l_attach: device %s is already known\n", devname);
	  return SANE_STATUS_GOOD;
	}
    }

  DBG (3, "v4l_attach: trying to open %s\n", devname);
  v4lfd = open (devname, O_RDWR);
  if (-1 != v4lfd)
    {
      if (ioctl (v4lfd, VIDIOCGCAP, &capability) < 0)
	{
	  DBG (1,
	       "sane_init: ioctl(%d, VIDIOCGCAP, ..) failed on `%s': %s\n",
	       v4lfd, devname, strerror (errno));
	  close (v4lfd);
	  return SANE_STATUS_INVAL;
	}
      if (!VID_TYPE_CAPTURE & capability.type)
	{
	  DBG (1,
	       "v4l_attach: device '%s' can't capture to memory -- exiting\n",
	       devname);
	  return SANE_STATUS_UNSUPPORTED;
	}
      DBG (2, "v4l_attach: found videodev '%s' on '%s'\n", capability.name,
	   devname);
      close (v4lfd);
    }
  else
    {
      DBG (1, "v4l_attach: failed to open device '%s' : %s\n", devname,
	   strerror (errno));
      return SANE_STATUS_INVAL;
    }

  dev = malloc (sizeof (*dev));
  if (!dev)
    return SANE_STATUS_NO_MEM;
  memset (dev, 0, sizeof (*dev));
  dev->sane.name = strdup (devname);
  if (!dev->sane.name)
    return SANE_STATUS_NO_MEM;
  dev->sane.model = strdup (capability.name);
  if (!dev->sane.model)
    return SANE_STATUS_NO_MEM;
  dev->sane.vendor = "Noname";
  dev->sane.type = "virtual device";

  ++num_devices;

  /* make this first of list */
  dev->next = first_dev;
  first_dev = dev;
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
SANE_Status
sane_init (SANE_Int * version_code, SANE_Auth_Callback authorize)
{
  char devname[PATH_MAX];
  SANE_String str;
  size_t len;
  FILE *fp;

  authorize = authorize;	/* stop gcc from complaining */
  DBG_INIT ();

  DBG (2, "sane v4l backend v%d.%d build %d from %s\n",
       V_MAJOR, V_MINOR, BUILD, PACKAGE_STRING);

  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, BUILD);

  fp = sanei_config_open (V4L_CONFIG_FILE);
  if (!fp)
    {
      DBG (2,
	   "sane_init: file `%s' not accessible (%s), trying /dev/video0\n",
	   V4L_CONFIG_FILE, strerror (errno));
      return v4l_attach ("/dev/video0");
    }

  while (sanei_config_read (devname, sizeof (devname), fp))
    {
      if (*devname == '#')
	continue;		/* ignore line comments */
      len = strlen (devname);
      if (!len)
	continue;		/* ignore empty lines */

      /* Remove trailing space and trailing comments */
      for (str = devname; *str && !isspace (*str) && *str != '#'; ++str);
      *str = '\0';
      v4l_attach (devname);
    }
  fclose (fp);
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
void
sane_exit (void)
{
  V4L_Device *dev, *next;

  for (dev = first_dev; dev; dev = next)
    {
      next = dev->next;
      if (dev->sane.name)
				free ((void *) dev->sane.name);
      if (dev->sane.model)
				free ((void *) dev->sane.model);
      free (dev);
    }
  if (devlist)
    {
      free (devlist);
      devlist = NULL;
    }
  DBG (5, "sane_exit: all devices freed\n");
}

/*----------------------------------------------------------------------------*/
SANE_Status
sane_get_devices (const SANE_Device *** device_list, SANE_Bool local_only)
{
  V4L_Device *dev;
  SANE_Int i;

  DBG (5, "sane_get_devices\n");
  local_only = SANE_TRUE;	/* Avoid compile warning */

  if (devlist)
    free (devlist);
  devlist = malloc ((num_devices + 1) * sizeof (devlist[0]));
  if (!devlist)
    return SANE_STATUS_NO_MEM;

  for (i = 0, dev = first_dev; i < num_devices; dev = dev->next, i++)
    {
      devlist[i] = &dev->sane;
    }
  devlist[i] = 0;

  *device_list = devlist;
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
SANE_Status
sane_open (SANE_String_Const devname, SANE_Handle * handle)
{
  V4L_Device *dev;
  V4L_Scanner *s;
  SANE_Int v4lfd;
  SANE_Status status;
  struct video_capability capability;

  v4lfd = -1;
  if (!devname)
    {
      DBG (1, "sane_open: devname == NULL\n");
      return SANE_STATUS_INVAL;
    }
  for (dev = first_dev; dev; dev = dev->next)
    {
      if (!strcmp (dev->sane.name, devname))
	{
	  DBG (5, "sane_open: device %s found in devlist\n", devname);
	  break;
	}
    }
  if (!devname[0])
    dev = first_dev;
  if (!dev)
    {
      DBG (1, "sane_open: device '%s' doesn't seem to be a v4l device\n",
	   devname);
      return SANE_STATUS_INVAL;
    }
  v4lfd = open (devname, O_RDWR);
  if (-1 == v4lfd)
    {
      DBG (1, "sane_open: can't open %s (%s)\n", devname, strerror (errno));
      return SANE_STATUS_DEVICE_BUSY;
    }
  if (ioctl (v4lfd, VIDIOCGCAP, &capability) < 0)
    {
      DBG (1, "sane_init: ioctl(%d, VIDIOCGCAP, ..) failed on `%s'\n",
	   v4lfd, devname);
      close (v4lfd);
      return SANE_STATUS_INVAL;
    }

  s = malloc (sizeof (*s));
  if (!s)
    return SANE_STATUS_NO_MEM;
  memset (s, 0, sizeof (*s));

  s->fd = v4lfd;

  status = v4l_init_options (s, capability.channels);
  if (status != SANE_STATUS_GOOD)
    return status;

  /* insert newly opened handle into list of open handles: */
  s->next = first_handle;
  first_handle = s;
  *handle = s;
  return status;
}

/*----------------------------------------------------------------------------*/
void
sane_close (SANE_Handle handle)
{
  V4L_Scanner *prev, *s;
  SANE_Int i;

  DBG (2, "sane_close: trying to close handle %p", (void *) handle);
  /* remove handle from list of open handles: */
  prev = NULL;
  for (s = first_handle; s; s = s->next)
    {
      if (s == handle)
	break;
      prev = s;
    }
  if (!s)
    {
      DBG (1, "sane_close: Invalid handle %p\n", handle);
      return;			/* oops, not a handle we know about */
    }

  /* Mute if possible */
  v4l_mute (s, SANE_TRUE);

  /* Free input selection strings */
  for (i = 0; s->InputNameArray[i]; i++)
    {
      if (s->InputNameArray[i])
	free (s->InputNameArray[i]);
    }

  /* Free channel string/frequency array */
  for (i = 0; i < s->NumChannels; i++)
    {
      if (s->Chnamearray[i])
	free (s->Chnamearray[i]);
    }
  if (s->NumChannels)
    {
      free (s->Chnamearray);
      free (s->Chfreqarray);
    }

  /* free current selected values */
  if (s->val[OPT_INPUT].s)
    free (s->val[OPT_INPUT].s);
  if (s->val[OPT_NORM].s)
    free (s->val[OPT_NORM].s);
  if (s->val[OPT_CHANNEL].s)
    free (s->val[OPT_CHANNEL].s);
  if (prev)
    prev->next = s->next;
  else
    first_handle = s->next;

  close (s->fd);
  free (s);
}

/*----------------------------------------------------------------------------*/
const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle handle, SANE_Int option)
{
  V4L_Scanner *s = handle;
  if (!s)
    {
      DBG (1, "sane_get_option_descriptor: Invalid handle (NULL\n");
      return NULL;
    }
  if ((unsigned) option >= NUM_OPTIONS || option < 0)
    {
      DBG (1, "sane_get_option_descriptor: Value option (%d) out of range\n",
	   option);
      return NULL;
    }
  DBG (4, "sane_get_option_descriptor: option %d (%s)\n", option,
       s->opt[option].name ? s->opt[option].name : s->opt[option].title);
  return s->opt + option;
}

/*----------------------------------------------------------------------------*/
SANE_Status
sane_control_option (SANE_Handle handle,
		     SANE_Int option,
		     SANE_Action action, void *val, SANE_Int * info)
{
  V4L_Scanner *s = handle;
  SANE_Status status;
  SANE_Word cap;
  struct video_picture pict;

  if (info)
    *info = 0;
  if (option >= NUM_OPTIONS || option < 0)
    {
      DBG (1, "sane_control_option: Value option (%d) out of range\n",
	   option);
      return SANE_STATUS_INVAL;
    }
  if (!s)
    {
      DBG (1, "sane_control_option: Invalid handle (NULL)\n");
      return SANE_STATUS_INVAL;
    }

  DBG (4, "sane_control_option: %s option %d (%s)\n",
       action == SANE_ACTION_GET_VALUE ? "get" :
       action == SANE_ACTION_SET_VALUE ? "set" :
       action == SANE_ACTION_SET_AUTO ? "auto set" :
       "(unknown action with)", option,
       s->opt[option].name ? s->opt[option].name : s->opt[option].title);

  cap = s->opt[option].cap;
  if (!SANE_OPTION_IS_ACTIVE (cap))
    {
      DBG (1, "sane_control_option: option is inactive\n");
      return SANE_STATUS_INVAL;
    }
  if (SANE_ACTION_GET_VALUE == action)
    {
      switch (option)
			  {
					/* word options: */
				case OPT_NUM_OPTS:
				case OPT_CHANNEL_UP:
				case OPT_CHANNEL_DOWN:
				case OPT_AUDIO:
				case OPT_DEPTH:
				case OPT_TL_X:
				case OPT_TL_Y:
				case OPT_BR_X:
				case OPT_BR_Y:
				case OPT_FINETUNE:
				case OPT_PRESET:
					*(SANE_Word *) val = s->val[option].w;
					return SANE_STATUS_GOOD;
					break;
					/* fixed options */
				case OPT_BRIGHTNESS:
				case OPT_CONTRAST:
				case OPT_HUE:
				case OPT_SATURATION:
					*(SANE_Fixed *) val = s->val[option].w;
					return SANE_STATUS_GOOD;
					break;
					/* string options */
				case OPT_NORM:
				case OPT_INPUT:
				case OPT_CHANNEL:
					if (s->val[option].s)
					  { 
							strcpy (val, s->val[option].s);
						}
					else
					  {
						  val = 0l;
							DBG (1, "sane_control_option: Option %d is not a string\n",
									 option);
						}
					return SANE_STATUS_GOOD;
				default:
						DBG (1, "sane_control_option: Option %d unknown\n", option);
				}
    }
  else if (action == SANE_ACTION_SET_VALUE)
    {
      if (!SANE_OPTION_IS_SETTABLE (s->opt[option].cap))
			  {
				  return SANE_STATUS_INVAL;
				}
      status = sanei_constrain_value (s->opt + option, val, info);
      if (status != SANE_STATUS_GOOD)
			  {
					DBG (1, "sane_control_option: sanei_constrain_value failed: %s\n",
							 sane_strstatus (status));
					return status;
				}
      if ((option >= OPT_BRIGHTNESS &&
					 option <= OPT_SATURATION) || option == OPT_DEPTH)
			  {
					if (ioctl (s->fd, VIDIOCGPICT, &pict) < 0)
					  {
							DBG (1,
									 "sane_control_option: Can not get Picture Parameters\n");
							return SANE_STATUS_IO_ERROR;
						}
				}
      switch (option)
			  {
				/* no-value options */
				case OPT_NUM_OPTS:
					break;
				/* (mostly) side-effect-free word options: */
				case OPT_TL_X:
				case OPT_TL_Y:
				case OPT_BR_X:
				case OPT_BR_Y:
					s->val[option].w = *(SANE_Word *) val;
					if (info)
						*info |= SANE_INFO_RELOAD_PARAMS;
				 break;
				 case OPT_DEPTH:
					 s->val[option].w = *(SANE_Word *) val;
					 pict.depth = *(SANE_Word *) val;
					 switch (pict.depth)
					   {
						 case 8:
							 pict.palette = VIDEO_PALETTE_GREY;
							 break;
						 case 16:
							 pict.palette = VIDEO_PALETTE_RGB565;
							 break;
						 case 24:
							 pict.palette = VIDEO_PALETTE_RGB24;
							 break;
						 }
					 if (info)
						 *info |= SANE_INFO_RELOAD_PARAMS;
					 break;
				 /* tuning related stuff */
				 case OPT_PRESET:
				 case OPT_FINETUNE:
					 s->val[option].w = *(SANE_Word *) val;
					 status = v4l_tune_channel (s);
					 if (SANE_STATUS_GOOD != status)
						 return status;
					 if (info && option == OPT_PRESET)
						 *info |= SANE_INFO_RELOAD_OPTIONS;
					 break;
				 case OPT_CHANNEL:
				 /* Show channel to user */
					 if (s->val[option].s)
						 free (s->val[option].s);
					 s->val[option].s = strdup (val);
					 if (!s->val[option].s)
						 return SANE_STATUS_NO_MEM;
					 break;
				 case OPT_CHANNEL_UP:
					 status = v4l_step_channel (s, UP);
					 if (SANE_STATUS_GOOD != status)
						 return status;
					 if (info)
						 *info |= SANE_INFO_RELOAD_OPTIONS;
					 break;
				 case OPT_CHANNEL_DOWN:
					 status = v4l_step_channel (s, DOWN);
					 if (SANE_STATUS_GOOD != status)
						 return status;
					 if (info)
						 *info |= SANE_INFO_RELOAD_OPTIONS;
					 break;
				 case OPT_AUDIO:
					 s->val[option].w = *(SANE_Word *) val;
					 v4l_mute (s, s->val[option].w != SANE_TRUE);
					 break;
				 case OPT_INPUT:
				   if (s->val[option].s)
						 free (s->val[option].s);
					 s->val[option].s = strdup (val);
					 if (!s->val[option].s)
						 return SANE_STATUS_NO_MEM;
					 status = v4l_set_input (s);
					 if (SANE_STATUS_GOOD != status)
						 return status;
					 if (info)
						 *info |= SANE_INFO_RELOAD_PARAMS | SANE_INFO_RELOAD_OPTIONS;
					 break;
				 case OPT_NORM:
					 if (s->val[option].s)
						 free (s->val[option].s);
					 s->val[option].s = strdup (val);
					 if (!s->val[option].s)
						 return SANE_STATUS_NO_MEM;
					 status = v4l_set_input (s);
					 if (SANE_STATUS_GOOD != status)
						 return status;
					 if (info)
						 *info |= SANE_INFO_RELOAD_PARAMS | SANE_INFO_RELOAD_OPTIONS;
 				   break;
				 case OPT_BRIGHTNESS:
					 s->val[option].w = *(SANE_Fixed *) val;
					 pict.brightness = v4l_scale_fix_percent2int (*(SANE_Fixed *) val);
					 break;
				 case OPT_CONTRAST:
					 s->val[option].w = *(SANE_Fixed *) val;
					 pict.contrast = v4l_scale_fix_percent2int (*(SANE_Fixed *) val);
				 break;
				 case OPT_HUE:
					 s->val[option].w = *(SANE_Fixed *) val;
					 pict.hue = v4l_scale_fix_percent2int (*(SANE_Fixed *) val);
					 break;
				 case OPT_SATURATION:
					 s->val[option].w = *(SANE_Fixed *) val;
					 pict.colour = v4l_scale_fix_percent2int (*(SANE_Fixed *) val);
					 break;
				 default:
						 DBG (1, "sane_control_option: option %d unknown\n", option);
						 return SANE_STATUS_INVAL;
				}
      if ((option >= OPT_BRIGHTNESS &&
					 option <= OPT_SATURATION) || option == OPT_DEPTH)
   			{
					if (ioctl (s->fd, VIDIOCSPICT, &pict) < 0)
						{
							DBG (1, "sane_control_option: ioctl VIDIOCAPICT failed (%s)\n",
									 strerror (errno));
							return SANE_STATUS_IO_ERROR;
						}
				}
      return SANE_STATUS_GOOD;
    }
  else if (action == SANE_ACTION_SET_AUTO)
    {
      if (!(cap & SANE_CAP_AUTOMATIC))
			  {
					DBG (1, "sane_control_option: option can't be set automatically\n");
					return SANE_STATUS_INVAL;
				}
      switch (option)
			  {
				case OPT_BRIGHTNESS:
				  /* not implemented yet */
				  return SANE_STATUS_GOOD;
			  default:
					break;
				}
    }
  return SANE_STATUS_INVAL;
}

/*----------------------------------------------------------------------------*/
SANE_Status
sane_get_parameters (SANE_Handle handle, SANE_Parameters * params)
{
  V4L_Scanner *s = handle;

  DBG (4, "sane_get_parameters\n");
  if (!s)
    {
      DBG (1, "sane_get_parameters: Invalid handle %p\n", handle);
      return SANE_STATUS_INVAL;
    }
  if (!params)
    {
      DBG (1, "sane_get_parameters: params == NULL\n");
      return SANE_STATUS_INVAL;
    }
  params->depth = 8;
  params->last_frame = SANE_TRUE;	/* stop here */
  params->lines = s->val[OPT_BR_Y].w - s->val[OPT_TL_Y].w;
  params->pixels_per_line = s->val[OPT_BR_X].w - s->val[OPT_TL_X].w;
  switch (s->val[OPT_DEPTH].w)
    {
    case 8:
      params->format = SANE_FRAME_GRAY;
      params->bytes_per_line = params->pixels_per_line;
      break;
    case 16:
    case 24:
      params->format = SANE_FRAME_RGB;
      params->bytes_per_line = params->pixels_per_line * 3;
      break;
    }
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
SANE_Status
sane_start (SANE_Handle handle)
{
  V4L_Scanner *s;
  struct video_mmap gb;
  struct video_mbuf ov_mbuf;
  struct video_capability capability;
  struct point_t windowsize;
  SANE_Int norm;

  DBG (2, "sane_start\n");
  for (s = first_handle; s; s = s->next)
    {
      if (s == handle)
	break;
    }
  if (!s)
    {
      DBG (1, "sane_start: Invalid handle (NULL)\n");
      return SANE_STATUS_INVAL;
    }

  /* Get Parms on the fly through the selected values */
  /* Capture the whole active area to the buffer */
  norm = v4l_get_option_int (s->val[OPT_NORM].s, Normlist, 0);
  windowsize.x = x_range[norm].min + x_range[norm].max + 1;
  windowsize.y = y_range[norm].min + y_range[norm].max + 1;
  if (ioctl (s->fd, VIDIOCGCAP, &capability) < 0)
    {
      DBG (1, "sane_start: Can not get capabilities\n");
      return SANE_STATUS_INVAL;	/* ??? IO_ERROR ??? */
    }
  if (ioctl (s->fd, VIDIOCGMBUF, &ov_mbuf) < 0)
    {
      DBG (1, "sane_start: No Framebuffer\n");
      return SANE_STATUS_INVAL;
    }
  DBG (2, "sane_start: Buffersize %d , Buffers %d , Offset %p\n",
       ov_mbuf.size, ov_mbuf.frames, ov_mbuf.offsets);
  s->Buffer =
    mmap (0, ov_mbuf.size, PROT_READ | PROT_WRITE, MAP_SHARED, s->fd, 0);
  gb.frame = 0;
  gb.width = windowsize.x;
  gb.height = windowsize.y;
  switch (s->val[OPT_DEPTH].w)
    {
		case 8:
			gb.format = VIDEO_PALETTE_GREY;
			break;
		case 16:
			gb.format = VIDEO_PALETTE_RGB565;
			break;
		case 24:
			gb.format = VIDEO_PALETTE_RGB24;
			break;
    }
  DBG (2, "sane_start: MMAPed Frame %d x %d with Format %d\n",
       gb.width, gb.height, gb.format);
  if (ioctl (s->fd, VIDIOCMCAPTURE, &gb) < 0)
    {
      DBG (1, "sane_start: ioctl(%d, VIDIOCMCAPTURE, ..) failed\n", s->fd);
      return SANE_STATUS_IO_ERROR;
    }
  DBG (2, "sane start: Waiting for Frame %x\n", gb.frame);
  if (ioctl (s->fd, VIDIOCSYNC, &(gb.frame)) < 0)
    {
      DBG (1, "sane_start: ioctl(%d, VIDIOCSYNC, ..) failed\n", s->fd);
      return SANE_STATUS_IO_ERROR;
    }
  s->ReadingPoint.x = s->val[OPT_TL_X].w;
  s->ReadingPoint.y = s->val[OPT_TL_Y].w;
  s->Pitch = windowsize.x * (s->val[OPT_DEPTH].w >> 3);
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
SANE_Status
sane_read (SANE_Handle handle,
	   SANE_Byte * buf, SANE_Int max_len, SANE_Int * lenp)
{
  SANE_Int offset;
  SANE_Int Bpp;
  SANE_Int px;
  V4L_Scanner *s = handle;

  if (!s)
    {
      DBG (1, "sane_read: Invalid handle (NULL)\n");
      return SANE_STATUS_INVAL;
    }
  *lenp = 0;
  if (s->ReadingPoint.x == s->val[OPT_BR_X].w &&
      s->ReadingPoint.y == s->val[OPT_BR_Y].w)
    {
      return SANE_STATUS_EOF;
    }
  DBG (3, "sane_read: transfering data from (%d,%d)\n",
       s->ReadingPoint.x, s->ReadingPoint.y);
  Bpp = s->val[OPT_DEPTH].w >> 3;
  if (s->val[OPT_DEPTH].w > 8)
    max_len = max_len - (max_len % 3);
  offset = s->ReadingPoint.y * s->Pitch;
  px = s->ReadingPoint.x * Bpp;
  while (*lenp < max_len)
    {
      switch (s->val[OPT_DEPTH].w)
			  {
				case 8:
					/* 8 bpp copy */
					*(buf + *lenp) = *(s->Buffer + offset + px);
					*lenp = *lenp + 1;
					break;
				case 16:
					/* 16 bpp copy */
					*(buf + *lenp) = *(s->Buffer + offset + px + 1) & 0xf8;
					*(buf + *lenp + 1) =
							(((*(s->Buffer + offset + px + 1)) << 5) +
							 (((*(s->Buffer + offset + px)) & 0x0e0) >> 3));
					*(buf + *lenp + 2) = *(s->Buffer + offset + px) << 3;
					*lenp = *lenp + 3;
					break;
				case 24:
  				/* swap red and blue (24 bpp copy / 32 bpp copy) */
					*(buf + *lenp) = *(s->Buffer + offset + px + 2);
					*(buf + *lenp + 1) = *(s->Buffer + offset + px + 1);
					*(buf + *lenp + 2) = *(s->Buffer + offset + px);
					*lenp = *lenp + 3;
					break;
				}
      px += Bpp;
      s->ReadingPoint.x++;
      if (s->ReadingPoint.x == s->val[OPT_BR_X].w)
			  {
					s->ReadingPoint.y++;
					if (s->ReadingPoint.y == s->val[OPT_BR_Y].w)
							break;
					s->ReadingPoint.x = s->val[OPT_TL_X].w;
					px = s->val[OPT_TL_X].w * Bpp;
					offset += s->Pitch;
			  }
    }
  DBG (3, "sane_read: to (%d,%d)\n", s->ReadingPoint.x, s->ReadingPoint.y);
  return SANE_STATUS_GOOD;
}

/*----------------------------------------------------------------------------*/
void
sane_cancel (SANE_Handle handle)
{
  DBG (2, "sane_cancel");
  handle = handle;
}


/*----------------------------------------------------------------------------*/
SANE_Status
sane_set_io_mode (SANE_Handle handle, SANE_Bool non_blocking)
{
  /* Avoid compile warning */
  handle = NULL;
  if (non_blocking == SANE_FALSE)
    return SANE_STATUS_GOOD;
  return SANE_STATUS_UNSUPPORTED;
}

/*----------------------------------------------------------------------------*/
SANE_Status
sane_get_select_fd (SANE_Handle handle, SANE_Int * fd)
{
  /* Avoid compile warning */
  handle = NULL;
  fd = NULL;
  return SANE_STATUS_UNSUPPORTED;
}