/* sane - Scanner Access Now Easy.
   Copyright (C) 1997 David Mosberger-Tang
   This file is part of the SANE package.
	 
   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.
	 
   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   Portions of this code are derived from Scott Laird's qcam driver.
   It's copyright notice is reproduced here:

   Copyright (C) 1996 by Scott Laird

   Permission is hereby granted, free of charge, to any person
   obtaining a copy of this software and associated documentation
   files (the "Software"), to deal in the Software without
   restriction, including without limitation the rights to use, copy,
   modify, merge, publish, distribute, sublicense, and/or sell copies
   of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be
   included in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
   NONINFRINGEMENT.  IN NO EVENT SHALL SCOTT LAIRD BE LIABLE FOR ANY
   CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
   CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */

#ifndef v4l_h
#define v4l_h

#include <sane/sane.h>

typedef enum
{
		OPT_NUM_OPTS = 0,
		
		OPT_MODE_GROUP,
		OPT_INPUT,			/* choose input */
		OPT_DEPTH,			/* 8 (b&w) or 24 (color) */
		OPT_NORM,			/* video norm */
		
		OPT_GEOMETRY_GROUP,
		OPT_TL_X,			/* top-left x */
		OPT_TL_Y,			/* top-left y */
		OPT_BR_X,			/* bottom-right x */
		OPT_BR_Y,			/* bottom-right y */
		
		OPT_ENHANCEMENT_GROUP,
		OPT_BRIGHTNESS,
		OPT_CONTRAST,
		OPT_HUE,
		OPT_SATURATION,
		
		OPT_SPECIAL_GROUP,
		OPT_AUDIO,			/* Switches Autio On/Off */
		OPT_CHANNEL,			/* select channel */
		OPT_PRESET,			/* Selected Preset */
		OPT_FINETUNE,			/* change received channel */
		OPT_CHANNEL_UP,
		OPT_CHANNEL_DOWN,		/* Up/Down Tuning buttons */
		/* must come last: */
		NUM_OPTIONS
}
V4L_Option;

typedef struct point_t
{
		SANE_Int x;
		SANE_Int y;
}
point_t;

typedef struct channellist_t
{
		struct channellist_t *next;
		unsigned long freq;
		SANE_String chname;
}
chanellist_t;

typedef union
{
		SANE_Word w;
		SANE_String s;
}
Option_Value;

typedef struct V4L_Device
{
		struct V4L_Device *next;
		SANE_Device sane;
}
V4L_Device;

typedef struct V4L_Scanner
{
		struct V4L_Scanner *next;
		
		SANE_Option_Descriptor opt[NUM_OPTIONS];
		Option_Value val[NUM_OPTIONS];
		SANE_Int fd;			/* Filedescriptor */
		/* strings for inputs */
		SANE_String *InputNameArray;
		/* state for reading a frame: */
		SANE_String Buffer;
		point_t ReadingPoint;
		SANE_Int Pitch;
		/* State of Tuner */
		unsigned long MinFrequency;
		unsigned long MaxFrequency;
		SANE_Int Scalefakt;
		/* freqencyarray for tuning */
		int NumChannels;
		unsigned long *Chfreqarray;
		SANE_String *Chnamearray;
}
V4L_Scanner;

#endif /* v4l_h */